/*===========================================================================



Arduino Lightning flash logger
==============================



Overview:
=========

CPU         :   Arduino UNO / ATmega328
Clock       :   XTAL 16MHz
Compiler    :   Arduino 1.0.5
Version     :   1.00
Date        :   2014-01-17
Author      :   Iacopo Giangrandi HB9DUL



Description:
============

This programs implements a logger for the lightning flash counter. It’s based 
on an Arduino UNO board, a data logging shield (SD-card plus RTC) and a 16x2 
LCD display. Each time a lightning strike is detected the date and time is 
recorded in a log file on an SD-card. The RTC has a battery backup and keep 
running even in case of power outage (frequent during storms). 

Events are represented by an optocoupler shorting for about 200ms each time a 
strike is detected. This triggers our logger via a digital input PIN.

Three LED signal the status of the logger. A green power LED connected on the 
power supply tells that the logger is powered up. A yellow LED connected on a 
digital pin echoes the input for checking the connection with the strike 
detector. This LED only works when the logger is ready to log. The yellow LED 
blinks in case of error telling that the logger is not functioning anymore. 
A red LED, connected on the SD-card chip select line, tells when the SD-card 
is accessed. This LED is connected via a PNP transistor, not to load too much 
this line.

When the logger is started, it looks on the root of SD-card for a file called 
SETTIME.CLK. If found, it will use it for adjusting the RTC. If the operation 
is successful, the file is deleted to prevent readjusting the clock on a 
wrong time if the logger is powered off and on again. See SETTIME.TXT for a 
file template and instruction.
Each time the logger starts, it will log it. Then, every time the input 
signal level drops (idle high, active low) the date and time are logged. Log 
file names are composed by the year and the month, 2014-01.LOG, for example. 
There is one log file per month and they are all located in the SD-card root 
directory. If nothing happens during one month, no file is created. All 
logged informations are also echoed on the virtual RS232 port over the USB 
for monitoring with a PC.

This logger cannot operate without SD-card. The SD-card must not be write 
protected and have sufficient space available.



I/O connections:
================

LCD connections:
----------------
D17/A3 ----> LCD_RS     (1)     LCD display Register Select line (0=command, 1=data)
D16/A2 ----> LCD_EN     (1)     LCD display Enable line (1=enable)
GND    ----- LCD_R/W    (1)     LCD display Read/Write line (0=write)
D7     ----> LCD_D4     (1)     LCD display data bit 4
D6     ----> LCD_D5     (1)     LCD display data bit 5
D5     ----> LCD_D6     (1)     LCD display data bit 6
D4     ----> LCD_D7     (1)     LCD display data bit 7

SD-card connections:
--------------------
D10    ----> SD_/CS     (2,3)   SD-card /chip select line, also used to switch on activity LED via a PNP transistor, not to load the line (1)
D11    ----> SD_MOSI    (2,3)   SD-card MOSI line
D12    <---- SD_MISO    (2,3)   SD-card MISO line
D13    ----> SD_SCK     (2,3)   SD-card SCK line

RTC connections:
----------------
A5/SCL ----> RTC_SCL    (2)     DS1307 I2C SCL line
A4/SDA <---> RTC_SDA    (2)     DS1307 I2C SDA line

UART connections:
-----------------
D0/RXD <---- USB_TX     (4)     UART communication (from PC to Arduino)
D1/TXD ----> USB_RX     (4)     UART communication (from Arduino to PC)

I/O connections:
----------------
D2     <---- IN_PIN     (1)     Logger input pin (1=idle, 0=active)
D3     ----> LED_PIN    (1)     LED output (1=on)

Notes:
------
(1)     Added on datalogger shield.
(2)     Standard on datalogger shield.
(3)     SD-card uses 3.3V levels, converter is standard on datalogger shield.
(4)     Standard on Arduino UNO



============================================================================= */





// --------------------------------------------------------------------------
// External files inclusions.
// --------------------------------------------------------------------------
#include <LiquidCrystal.h>                                      // Include dot-matrix LCD library.
#include <SD.h>                                                 // Iclude SD-card library.
#include <Wire.h>                                               // Include I2C library.
#include "RTClib.h"                                             // Include DS1307 Real Time Clock library. This library is external and has to be manually installed.



// --------------------------------------------------------------------------
// General definitions, constants and global variables.
// --------------------------------------------------------------------------
#define       SET_TIME_FILE     "SETTIME.CLK"                   // File name used to set the RTC.
#define       LOG_FILE_EXT      ".LOG"                          // File extension used for log files (name is the year followed by the month, like "2014-01.LOG".



#define       LOG_MSG_START     "Log session started."          // Message appended to the log each time the logger is powered up and ready to log.
#define       LOG_MSG           "Lightning strike detected."    // Message appended to the log each time a ligthning strike is detected.

#define       ERR_SD_CARD       "SD card error."                // Message shown on the display in case of error with the SD-card.
#define       WRN_CLK_STOP_MSG  "RTC NOT running."              // Message shown on the display when the RTC is not running. Just set the clokc via the SD-card to start it.
#define       CLK_UPD_MSG       "Clock adjusted."               // Message shown on the display when the RTC has been adjusted to a new time/date.
#define       WELCOME_MSG_1     "Lightning  flash"              // Message shown on the display each time the logger is powered up with no errors (line 1).
#define       WELCOME_MSG_2     "counter - HB9DUL"              // Message shown on the display each time the logger is powered up with no errors (line 2).

int global_count = 0;                                           // A global counter for all lightnings detected. This variable is not saved on the SD-card and is reset to 0 every time the logger is powered up.



// --------------------------------------------------------------------------
// Hardware related definitions, constants and global variables.
// --------------------------------------------------------------------------
#define       IN_PIN            2                               // Logger input connected to digital pin 2, idle high, active low.
#define       LED_PIN           3                               // LED connected to digital pin 3, high=on.

LiquidCrystal lcd(17, 16, 7, 6, 5, 4);                          // Initialize the library with pin numbers and create "lcd" object (RS=pin17, EN=pin16, (R/W=gnd), D4=pin7, D5=pin6, D6=pin5, D7=pin4).
byte lightning_char[8] =                                        // A charcater pattern for a nice lightning symbol.
{
    B00010,
    B00100,
    B01111,
    B00010,
    B10100,
    B11000,
    B11110,
};

const int chipSelect = 10;                                      // For the SD-card, CS is pin 10. This pin is also used to monitor SD-card activity on a LED.

RTC_DS1307 rtc;                                                 // Define "rtc" as a RTC_DS1307 object.



// --------------------------------------------------------------------------
// This is the setup procedure executed once at startup. This procedure is
// automatically executed by Arduino.
// --------------------------------------------------------------------------
void setup()
{

    // Configure I/O lines.

    pinMode(IN_PIN, INPUT);                                     // Digital pin 2 is the logger input.
    pinMode(LED_PIN, OUTPUT);                                   // Digital pin 3 is an output for the LED.



    // Configure and initialize LCD display.

    lcd.begin(16, 2);                                           // Set up the LCD's display (16 columns by 2 rows).
    lcd.createChar(0, lightning_char);                          // Set character 0 as the "lightning" character.



    // Configure serial port.

    Serial.begin(9600);                                         // Set serial communication at 9600N81.



    // Initialeze SD-card.

    pinMode(chipSelect, OUTPUT);                                // Make sure that chip select pin is set to output.
    if (!SD.begin(chipSelect))                                  // Check if the card is present and can be initialized.
        error_msg(ERR_SD_CARD);                                 // Stop everything if it's not the case.



    // Initialize RTC.

    Wire.begin();                                               // Initialize I2C bus first.
    rtc.begin();                                                // Than initalize RTC.
    if (!rtc.isrunning())                                       // Warn if RTC is not running.
        Serial.println(WRN_CLK_STOP_MSG);                       // If not running, the message is only sent on the serial port, since it's evident on the display that the clock is frozen... Just set the clock to start it.
    set_clock();                                                // This procedure will check on the SD-card if there is a file with a new time to adjust the clock.



    // Now we are done with initializations, we can show the welocme message.

    lcd.clear();                                                // Clear the screen (previous errors or messages could be currently on the display).
    lcd.print(WELCOME_MSG_1);                                   // Print first welcome line.
    lcd.setCursor(0, 1);                                        // Move to line 2.
    lcd.print(WELCOME_MSG_2);                                   // Print second welcome line.

    delay(2000);                                                // Wait 2 seconds.

    lcd.clear();                                                // Clear the screen again.



    // Start logging.

    log_event(LOG_MSG_START);                                   // Log on the SD-card that the logger has started.

    global_count = 0;                                           // Set global counter to 0.
    display_count();                                            // And display it.

}



// --------------------------------------------------------------------------
// This is the main program. It's executed automatically after the 
// initialization in an andless loop by Arduino.
// --------------------------------------------------------------------------
void loop()
{
    if (digitalRead(IN_PIN) == 0)                               // Check the status of the input line: when this line drops a lightning is detected.
    {
        digitalWrite(LED_PIN, 1);                               // Switch the LED on to show this condition.

        global_count++;                                         // Increase global counter.
        display_count();                                        // And display it.

        log_event(LOG_MSG);                                     // Log the event on the SD-card.

        while (digitalRead(IN_PIN) == 0);                       // Wait until input is idle again.
        delay(50);                                              // Debounce for 50ms.

        digitalWrite(LED_PIN, 0);                               // Switch the LED back off.
    }

    display_clock();                                            // When idle, display the current date and time.
    delay(50);                                                  // Wait for 50ms, not to update the disply too often (even if it wouldn't be a problem).
}



// --------------------------------------------------------------------------
// This procedure (called in case of error) displays the passed message on
// the LCD and on the serial port, than blinks the LED at 1Hz. It never
// returns.
// --------------------------------------------------------------------------
void error_msg(char message[])
{
    lcd.clear();                                                // Clear LCD display.
    lcd.print(message);                                         // Write the message.
    Serial.println(message);                                    // Write it also on the serial port.

    while (1)                                                   // Infinitie loop that blinks the LED at 1Hz (roughly).
    {
        digitalWrite(LED_PIN, 1);
        delay(500);
        digitalWrite(LED_PIN, 0);
        delay(500);
    }
}



// --------------------------------------------------------------------------
// This procedure checks for the presence of the "SETTIME.CLK" file on the 
// SD-card. If found (and large enough), the first 19 characters are read
// and used to set the clock. File format is:
// 
// YYYY-MM-DD HH:MM:SS
// 
// Only the beginning of the file is read, all the rest is ignored. The 
// position of the digits is important, use leading zeros. Separators are 
// ignored. No consistency check is performed.
// --------------------------------------------------------------------------
void set_clock()
{
    uint16_t set_year;                                          // 6 variables used to store the date/time information.
    uint8_t set_month;
    uint8_t set_day;
    uint8_t set_hour;
    uint8_t set_minute;
    uint8_t set_second;

    if (SD.exists(SET_TIME_FILE))                               // Check if "SETTIME.CLK" exists.
    {
        File dataFile = SD.open(SET_TIME_FILE);                 // Yes: open it.
        if (dataFile)                                           // Check if operation is successful.
        {
            if (dataFile.available() >= 19)                     // Check if the file is large enough.
            {
                set_year = (dataFile.read() - '0') * 1000;      // Read four digits of the year and convert to an integer.
                set_year += (dataFile.read() - '0') * 100;
                set_year += (dataFile.read() - '0') * 10;
                set_year += dataFile.read() - '0';

                dataFile.read();                                // Ignore separator.

                set_month = (dataFile.read() - '0') * 10;       // Read two digits of the month and convert to an integer.
                set_month += dataFile.read() - '0';

                dataFile.read();                                // Ignore separator.

                set_day = (dataFile.read() - '0') * 10;         // Read two digits of the day and convert to an integer.
                set_day += dataFile.read() - '0';

                dataFile.read();                                // Ignore separator.

                set_hour = (dataFile.read() - '0') * 10;        // Read two digits of the hour and convert to an integer.
                set_hour += dataFile.read() - '0';

                dataFile.read();                                // Ignore separator.

                set_minute = (dataFile.read() - '0') * 10;      // Read two digits of the minutes and convert to an integer.
                set_minute += dataFile.read() - '0';

                dataFile.read();                                // Ignore separator.

                set_second = (dataFile.read() - '0') * 10;      // Read two digits of the seconds and convert to an integer.
                set_second += dataFile.read() - '0';

                dataFile.close();                               // Done with this file: close it.

                rtc.adjust(DateTime(set_year, set_month, set_day, set_hour, set_minute, set_second));   // Now we can adjust the clock with the collected data.

                SD.remove(SET_TIME_FILE);                       // Once the clock is set, delete the file, so that the clock wont be set again if the logger is restarted.

                Serial.print(CLK_UPD_MSG);                      // Tell that the clock has been updated on the serial line.
                lcd.clear();                                    // And on the display as well.
                lcd.print(CLK_UPD_MSG);
                delay(1000);                                    // Wait one second to allow reading the message.
            }
        }
        else
            error_msg(ERR_SD_CARD);                             // Something went wrong opening the file, thorw an error.
    }
}



// --------------------------------------------------------------------------
// This procedure shows the current date and time on the display.
// --------------------------------------------------------------------------
void display_clock()
{
    DateTime now = rtc.now();                                   // Get current date and time.

    lcd.setCursor(0, 0);                                        // Go at the beginning of line 1.
    lcd.print(now.year(), DEC);                                 // Write the year.

    lcd.print('-');                                             // Write the separator.

    if (now.month() < 10)                                       // Write the month with leading zero.
        lcd.print('0');
    lcd.print(now.month(), DEC);

    lcd.print('-');                                             // Write the separator.

    if (now.day() < 10)                                         // Write the day with leading zero.
        lcd.print('0');
    lcd.print(now.day(), DEC);

    lcd.setCursor(0, 1);                                        // Go at the beginning of line 2.

    if (now.hour() < 10)                                        // Write the hour with leading zero.
        lcd.print('0');
    lcd.print(now.hour(), DEC);

    lcd.print(':');                                             // Write the separator.

    if (now.minute() < 10)                                      // Write the minute with leading zero.
        lcd.print('0');
    lcd.print(now.minute(), DEC);

    lcd.print(':');                                             // Write the separator.

    if (now.second() < 10)                                      // Write the second with leading zero.
        lcd.print('0');
    lcd.print(now.second(), DEC);
}



// --------------------------------------------------------------------------
// This procedure displays the global counter and a nice "lightning 
// animation" to show the logger is working.
// --------------------------------------------------------------------------
void display_count()
{
    byte i, j;                                                  // Two dummy variables.

    lcd.setCursor(11, 1);                                       // Go at the "end" of line 2.
    if (global_count < 10000)                                   // Write counter, aligned on the right.
        lcd.print(' ');
    if (global_count < 1000)
        lcd.print(' ');
    if (global_count < 100)
        lcd.print(' ');
    if (global_count < 10)
        lcd.print(' ');
    lcd.print(global_count, DEC);

    lcd.setCursor(12, 0);                                       // Go at the end of line 1.
    if (global_count == 0)                                      // If counter is still 0, just display and empty "----".
        lcd.print("----");
    else
    {
        j = (global_count + 3) % 4;                             // Diplay a running lightning (character '\0') shifted every time a lightning is detected. The "+3" is to have the first ligthing (global counter at 1) in the first position instead of the second.
        for (i = 0; i < 4; i++)                                 // Write a string fo 4 character with a rinning lightning, from left to right, something like "/---", "-/--", "--/-", "---/" and so on.
            if (i == j)
                lcd.print('\0');
            else
                lcd.print('-');
    }
}



// --------------------------------------------------------------------------
// This is the procedure that logs the events on the SD-card (en echos them
// also on the serial port). The passed "event_text" is appended after the
// time stamp and the global counter.
// --------------------------------------------------------------------------
void log_event(char event_text[])
{
    DateTime now = rtc.now();                                   // Get current date and time.

    String temp;                                                // "temp" is a string object to create the name of the log file.
    temp = String(now.year(), DEC);                             // Convert the year into a string: this is the beginning of the log file name.
    temp += '-';                                                // Add a '-' character.
    if (now.month() < 10)                                       // Add the month, with leading zero.
        temp += '0';
    temp += String(now.month(), DEC);
    temp += LOG_FILE_EXT;                                       // Add the extension. Now the file name should look something like "2014-01.LOG".
    char log_file_name[temp.length()+1];                        // Unfortunately, SD-card procedure only accept character arrays and not string objects: so, let's create a character array "log_file_name", of the same size of the above string.
    temp.toCharArray(log_file_name, sizeof(log_file_name));     // Let's now convert the string into a character array.

    File dataFile = SD.open(log_file_name, FILE_WRITE);         // Open the log file for writing (default action is "append").
    if (dataFile)                                               // If the operation is successful, write to it:
    {
        dataFile.print(now.year(), DEC);                        // Write the date to the log file.

        dataFile.print('-');

        if (now.month() < 10)
            dataFile.print('0');
        dataFile.print(now.month(), DEC);

        dataFile.print('-');

        if (now.day() < 10)
            dataFile.print('0');
        dataFile.print(now.day(), DEC);

        dataFile.print(' ');

        if (now.hour() < 10)                                    // Write the time to the log file.
            dataFile.print('0');
        dataFile.print(now.hour(), DEC);

        dataFile.print(':');

        if (now.minute() < 10)
            dataFile.print('0');
        dataFile.print(now.minute(), DEC);

        dataFile.print(':');

        if (now.second() < 10)
            dataFile.print('0');
        dataFile.print(now.second(), DEC);

        dataFile.print(" - ");                                  // Separate with a " - ".

        dataFile.print(global_count, DEC);                      // Write the global counter (no padding nor leading zeros).

        dataFile.print(" - ");                                  // Separate with a " - ".

        dataFile.println(event_text);                           // Write the event text string and terminate the line.

        dataFile.close();                                       // We are done with the file write operation, close it.
    }
    else 
        error_msg(ERR_SD_CARD);                                 // Something went wrong with the file, throw an error.

    Serial.print(now.year(), DEC);                              // Write the same logging information on the serial line.

    Serial.print('-');

    if (now.month() < 10)
        Serial.print('0');
    Serial.print(now.month(), DEC);

    Serial.print('-');

    if (now.day() < 10)
        Serial.print('0');
    Serial.print(now.day(), DEC);

    Serial.print(' ');

    if (now.hour() < 10)
        Serial.print('0');
    Serial.print(now.hour(), DEC);

    Serial.print(':');

    if (now.minute() < 10)
        Serial.print('0');
    Serial.print(now.minute(), DEC);

    Serial.print(':');

    if (now.second() < 10)
        Serial.print('0');
    Serial.print(now.second(), DEC);

    Serial.print(" - ");

    Serial.print(global_count, DEC);

    Serial.print(" - ");

    Serial.println(event_text);
}



